@@ -16,8 +16,9 @@ const maybeReadFile = file => {
   try {
     return fs.readFileSync(file, 'utf8')
   } catch (er) {
-    if (er.code !== 'ENOENT')
+    if (er.code !== 'ENOENT') {
       throw er
+    }
     return null
   }
 }
@@ -27,27 +28,32 @@ const buildOmitList = obj => {
   const omit = obj.omit || []
 
   const only = obj.only
-  if (/^prod(uction)?$/.test(only) || obj.production)
+  if (/^prod(uction)?$/.test(only) || obj.production) {
     omit.push('dev')
-  else if (obj.production === false)
+  } else if (obj.production === false) {
     include.push('dev')
+  }
 
-  if (/^dev/.test(obj.also))
+  if (/^dev/.test(obj.also)) {
     include.push('dev')
+  }
 
-  if (obj.dev)
+  if (obj.dev) {
     include.push('dev')
+  }
 
-  if (obj.optional === false)
+  if (obj.optional === false) {
     omit.push('optional')
-  else if (obj.optional === true)
+  } else if (obj.optional === true) {
     include.push('optional')
+  }
 
   obj.omit = [...new Set(omit)].filter(type => !include.includes(type))
   obj.include = [...new Set(include)]
 
-  if (obj.omit.includes('dev'))
+  if (obj.omit.includes('dev')) {
     process.env.NODE_ENV = 'production'
+  }
 
   return obj.omit
 }
@@ -98,8 +104,9 @@ const {
 
 const define = (key, def) => {
   /* istanbul ignore if - this should never happen, prevents mistakes below */
-  if (definitions[key])
+  if (definitions[key]) {
     throw new Error(`defining key more than once: ${key}`)
+  }
   definitions[key] = new Definition(key, def)
 }
 
@@ -160,6 +167,12 @@ define('access', {
     (and installable) set \`--access=public\`. The only valid values for
     \`access\` are \`public\` and \`restricted\`. Unscoped packages _always_
     have an access level of \`public\`.
+
+    Note: Using the \`--access\` flag on the \`npm publish\` command will only
+    set the package access level on the initial publish of the package. Any
+    subsequent \`npm publish\` commands using the \`--access\` flag will not
+    have an effect to the access level.  To make changes to the access level
+    after the initial publish use \`npm access\`.
   `,
   flatten,
 })
@@ -336,8 +349,9 @@ define('cache-max', {
     This option has been deprecated in favor of \`--prefer-online\`
   `,
   flatten (key, obj, flatOptions) {
-    if (obj[key] <= 0)
+    if (obj[key] <= 0) {
       flatOptions.preferOnline = true
+    }
   },
 })
 
@@ -351,8 +365,9 @@ define('cache-min', {
     This option has been deprecated in favor of \`--prefer-offline\`.
   `,
   flatten (key, obj, flatOptions) {
-    if (obj[key] >= 9999)
+    if (obj[key] >= 9999) {
       flatOptions.preferOffline = true
+    }
   },
 })
 
@@ -366,12 +381,14 @@ define('cafile', {
   `,
   flatten (key, obj, flatOptions) {
     // always set to null in defaults
-    if (!obj.cafile)
+    if (!obj.cafile) {
       return
+    }
 
     const raw = maybeReadFile(obj.cafile)
-    if (!raw)
+    if (!raw) {
       return
+    }
 
     const delim = '-----END CERTIFICATE-----'
     flatOptions.ca = raw.replace(/\r\n/g, '\n').split(delim)
@@ -455,7 +472,10 @@ define('color', {
   flatten (key, obj, flatOptions) {
     flatOptions.color = !obj.color ? false
       : obj.color === 'always' ? true
-      : process.stdout.isTTY
+      : !!process.stdout.isTTY
+    flatOptions.logColor = !obj.color ? false
+      : obj.color === 'always' ? true
+      : !!process.stderr.isTTY
   },
 })
 
@@ -798,7 +818,12 @@ define('global', {
     * bin files are linked to \`{prefix}/bin\`
     * man pages are linked to \`{prefix}/share/man\`
   `,
-  flatten,
+  flatten: (key, obj, flatOptions) => {
+    flatten(key, obj, flatOptions)
+    if (flatOptions.global) {
+      flatOptions.location = 'global'
+    }
+  },
 })
 
 define('global-style', {
@@ -908,6 +933,19 @@ define('include-staged', {
   flatten,
 })
 
+define('include-workspace-root', {
+  default: false,
+  type: Boolean,
+  description: `
+    Include the workspace root when workspaces are enabled for a command.
+
+    When false, specifying individual workspaces via the \`workspace\` config,
+    or all workspaces via the \`workspaces\` flag, will cause npm to operate only
+    on the specified workspaces, and not on the root project.
+  `,
+  flatten,
+})
+
 define('init-author-email', {
   default: '',
   type: String,
@@ -1125,14 +1163,40 @@ define('location', {
   description: `
     When passed to \`npm config\` this refers to which config file to use.
   `,
-  // NOTE: the flattener here deliberately does not alter the value of global
-  // for now, this is to avoid inadvertently causing any breakage. the value of
-  // global, however, does modify this flag.
-  flatten (key, obj, flatOptions) {
-    // if global is set, we override ourselves
-    if (obj.global)
-      obj.location = 'global'
-    flatOptions.location = obj.location
+  flatten: (key, obj, flatOptions) => {
+    flatten(key, obj, flatOptions)
+    if (flatOptions.global) {
+      flatOptions.location = 'global'
+    }
+  },
+})
+
+define('lockfile-version', {
+  default: null,
+  type: [null, 1, 2, 3, '1', '2', '3'],
+  defaultDescription: `
+    Version 2 if no lockfile or current lockfile version less than or equal to
+    2, otherwise maintain current lockfile version
+  `,
+  description: `
+    Set the lockfile format version to be used in package-lock.json and
+    npm-shrinkwrap-json files.  Possible options are:
+
+    1: The lockfile version used by npm versions 5 and 6.  Lacks some data that
+    is used during the install, resulting in slower and possibly less
+    deterministic installs.  Prevents lockfile churn when interoperating with
+    older npm versions.
+
+    2: The default lockfile version used by npm version 7.  Includes both the
+    version 1 lockfile data and version 3 lockfile data, for maximum
+    determinism and interoperability, at the expense of more bytes on disk.
+
+    3: Only the new lockfile information introduced in npm version 7.  Smaller
+    on disk than lockfile version 2, but not interoperable with older npm
+    versions.  Ideal if all users are on npm version 7 and higher.
+  `,
+  flatten: (key, obj, flatOptions) => {
+    flatOptions.lockfileVersion = obj[key] && parseInt(obj[key], 10)
   },
 })
 
@@ -1150,8 +1214,8 @@ define('loglevel', {
     'silly',
   ],
   description: `
-    What level of logs to report.  On failure, *all* logs are written to
-    \`npm-debug.log\` in the current working directory.
+    What level of logs to report.  All logs are written to a debug log,
+    with the path to that file printed if the execution of a command fails.
 
     Any logs of a higher level than the setting are shown. The default is
     "notice".
@@ -1233,10 +1297,11 @@ define('noproxy', {
     Also accepts a comma-delimited string.
   `,
   flatten (key, obj, flatOptions) {
-    if (Array.isArray(obj[key]))
+    if (Array.isArray(obj[key])) {
       flatOptions.noProxy = obj[key].join(',')
-    else
+    } else {
       flatOptions.noProxy = obj[key]
+    }
   },
 })
 
@@ -1353,7 +1418,12 @@ define('package-lock', {
     modules will also be disabled.  To remove extraneous modules with
     package-locks disabled use \`npm prune\`.
   `,
-  flatten,
+  flatten: (key, obj, flatOptions) => {
+    flatten(key, obj, flatOptions)
+    if (flatOptions.packageLockOnly) {
+      flatOptions.packageLock = true
+    }
+  },
 })
 
 define('package-lock-only', {
@@ -1369,7 +1439,12 @@ define('package-lock-only', {
     For \`list\` this means the output will be based on the tree described by the
     \`package-lock.json\`, rather than the contents of \`node_modules\`.
   `,
-  flatten,
+  flatten: (key, obj, flatOptions) => {
+    flatten(key, obj, flatOptions)
+    if (flatOptions.packageLockOnly) {
+      flatOptions.packageLock = true
+    }
+  },
 })
 
 define('pack-destination', {
@@ -1461,6 +1536,10 @@ define('progress', {
 
     Set to \`false\` to suppress the progress bar.
   `,
+  flatten (key, obj, flatOptions) {
+    flatOptions.progress = !obj.progress ? false
+      : !!process.stderr.isTTY && process.env.TERM !== 'dumb'
+  },
 })
 
 define('proxy', {
@@ -1504,14 +1583,18 @@ define('registry', {
 
 define('save', {
   default: true,
+  defaultDescription: `\`true\` unless when using \`npm update\` or
+  \`npm dedupe\` where it defaults to \`false\``,
   usage: '-S|--save|--no-save|--save-prod|--save-dev|--save-optional|--save-peer',
   type: Boolean,
   short: 'S',
   description: `
-    Save installed packages to a package.json file as dependencies.
+    Save installed packages to a \`package.json\` file as dependencies.
 
     When used with the \`npm rm\` command, removes the dependency from
-    package.json.
+    \`package.json\`.
+
+    Will also prevent writing to \`package-lock.json\` if set to \`false\`.
   `,
   flatten,
 })
@@ -1551,8 +1634,9 @@ define('save-dev', {
   `,
   flatten (key, obj, flatOptions) {
     if (!obj[key]) {
-      if (flatOptions.saveType === 'dev')
+      if (flatOptions.saveType === 'dev') {
         delete flatOptions.saveType
+      }
       return
     }
 
@@ -1584,20 +1668,23 @@ define('save-optional', {
   `,
   flatten (key, obj, flatOptions) {
     if (!obj[key]) {
-      if (flatOptions.saveType === 'optional')
+      if (flatOptions.saveType === 'optional') {
         delete flatOptions.saveType
-      else if (flatOptions.saveType === 'peerOptional')
+      } else if (flatOptions.saveType === 'peerOptional') {
         flatOptions.saveType = 'peer'
+      }
       return
     }
 
-    if (flatOptions.saveType === 'peerOptional')
+    if (flatOptions.saveType === 'peerOptional') {
       return
+    }
 
-    if (flatOptions.saveType === 'peer')
+    if (flatOptions.saveType === 'peer') {
       flatOptions.saveType = 'peerOptional'
-    else
+    } else {
       flatOptions.saveType = 'optional'
+    }
   },
 })
 
@@ -1605,24 +1692,27 @@ define('save-peer', {
   default: false,
   type: Boolean,
   description: `
-    Save installed packages. to a package.json file as \`peerDependencies\`
+    Save installed packages to a package.json file as \`peerDependencies\`
   `,
   flatten (key, obj, flatOptions) {
     if (!obj[key]) {
-      if (flatOptions.saveType === 'peer')
+      if (flatOptions.saveType === 'peer') {
         delete flatOptions.saveType
-      else if (flatOptions.saveType === 'peerOptional')
+      } else if (flatOptions.saveType === 'peerOptional') {
         flatOptions.saveType = 'optional'
+      }
       return
     }
 
-    if (flatOptions.saveType === 'peerOptional')
+    if (flatOptions.saveType === 'peerOptional') {
       return
+    }
 
-    if (flatOptions.saveType === 'optional')
+    if (flatOptions.saveType === 'optional') {
       flatOptions.saveType = 'peerOptional'
-    else
+    } else {
       flatOptions.saveType = 'peer'
+    }
   },
 })
 
@@ -1659,8 +1749,9 @@ define('save-prod', {
   `,
   flatten (key, obj, flatOptions) {
     if (!obj[key]) {
-      if (flatOptions.saveType === 'prod')
+      if (flatOptions.saveType === 'prod') {
         delete flatOptions.saveType
+      }
       return
     }
 
@@ -1702,7 +1793,10 @@ define('scope', {
   `,
   flatten (key, obj, flatOptions) {
     const value = obj[key]
-    flatOptions.projectScope = value && !/^@/.test(value) ? `@${value}` : value
+    const scope = value && !/^@/.test(value) ? `@${value}` : value
+    flatOptions.scope = scope
+    // projectScope is kept for compatibility with npm-registry-fetch
+    flatOptions.projectScope = scope
   },
 })
 
@@ -2029,8 +2123,9 @@ define('user-agent', {
     const value = obj[key]
     const ciName = obj['ci-name']
     let inWorkspaces = false
-    if (obj.workspaces || obj.workspace && obj.workspace.length)
+    if (obj.workspaces || obj.workspace && obj.workspace.length) {
       inWorkspaces = true
+    }
     flatOptions.userAgent =
       value.replace(/\{node-version\}/gi, obj['node-version'])
         .replace(/\{npm-version\}/gi, obj['npm-version'])
@@ -2039,10 +2134,14 @@ define('user-agent', {
         .replace(/\{workspaces\}/gi, inWorkspaces)
         .replace(/\{ci\}/gi, ciName ? `ci/${ciName}` : '')
         .trim()
+
+    // We can't clobber the original or else subsequent flattening will fail
+    // (i.e. when we change the underlying config values)
+    // obj[key] = flatOptions.userAgent
+
     // user-agent is a unique kind of config item that gets set from a template
     // and ends up translated.  Because of this, the normal "should we set this
     // to process.env also doesn't work
-    obj[key] = flatOptions.userAgent
     process.env.npm_config_user_agent = flatOptions.userAgent
   },
 })
@@ -2119,24 +2218,48 @@ define('workspace', {
 
     * Workspace names
     * Path to a workspace directory
-    * Path to a parent workspace directory (will result to selecting all of the
-      nested workspaces)
+    * Path to a parent workspace directory (will result in selecting all
+      workspaces within that folder)
 
     When set for the \`npm init\` command, this may be set to the folder of
     a workspace which does not yet exist, to create the folder and set it
     up as a brand new workspace within the project.
   `,
+  flatten: (key, obj, flatOptions) => {
+    definitions['user-agent'].flatten('user-agent', obj, flatOptions)
+  },
 })
 
 define('workspaces', {
-  default: false,
-  type: Boolean,
+  default: null,
+  type: [null, Boolean],
   short: 'ws',
   envExport: false,
   description: `
-    Enable running a command in the context of **all** the configured
+    Set to true to run the command in the context of **all** configured
     workspaces.
+
+    Explicitly setting this to false will cause commands like \`install\` to
+    ignore workspaces altogether.
+    When not set explicitly:
+
+    - Commands that operate on the \`node_modules\` tree (install, update,
+      etc.) will link workspaces into the \`node_modules\` folder.
+    - Commands that do other things (test, exec, publish, etc.) will operate
+      on the root project, _unless_ one or more workspaces are specified in
+      the \`workspace\` config.
   `,
+  flatten: (key, obj, flatOptions) => {
+    definitions['user-agent'].flatten('user-agent', obj, flatOptions)
+
+    // TODO: this is a derived value, and should be reworked when we have a
+    // pattern for derived value
+
+    // workspacesEnabled is true whether workspaces is null or true
+    // commands contextually work with workspaces or not regardless of
+    // configuration, so we need an option specifically to disable workspaces
+    flatOptions.workspacesEnabled = obj[key] !== false
+  },
 })
 
 define('yes', {
